/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation;

import java.util.Arrays;
import java.util.UUID;
import java.util.stream.Stream;

import org.apache.logging.log4j.ThreadContext;
import org.springframework.beans.factory.annotation.Lookup;
import org.springframework.boot.CommandLineRunner;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

import de.ozgcloud.aggregation.transformation.AggregationMapping;
import de.ozgcloud.aggregation.transformation.Transformation;
import de.ozgcloud.aggregation.transformation.TransformationException;
import de.ozgcloud.aggregation.transformation.TransformationService;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Component
@RequiredArgsConstructor
@Log4j2
@ConditionalOnProperty(prefix = "ozgcloud.command.line.runner", value = "enabled", havingValue = "true", matchIfMissing = true)
public abstract class AggregationManagerRunner implements CommandLineRunner {

	private static final String MDC_EXECUTION = "execution";

	private final AggregationManagerProperties aggregationManagerProperties;
	private final TransformationProperties transformationProperties;
	private final TransformationService transformationService;
	private final AggregationDataLoaderRegistry loaderRegistry;

	@Override
	public void run(String... args) throws TransformationException {
		getScopesWithoutConfiguredTransformations().forEach(this::runWithDefaultTransformation);
		runWithConfiguredTransformations();
	}

	Stream<AggregationMapping.Scope> getScopesWithoutConfiguredTransformations() {
		return Arrays.stream(AggregationMapping.Scope.values()).filter(loaderRegistry::hasLoader).filter(this::hasNoConfiguredTransformations);
	}

	boolean hasNoConfiguredTransformations(AggregationMapping.Scope scope) {
		return transformationProperties.getAggregationMappings().stream().map(AggregationMapping::getScope).noneMatch(scope::equals);
	}

	void runWithDefaultTransformation(AggregationMapping.Scope scope) throws TransformationException {
		runWithTransformation(
				transformationService.load(transformationProperties.getIdentifier(), null),
				createAggregator(scope));
	}

	void runWithConfiguredTransformations() {
		transformationProperties.getAggregationMappings().forEach(aggregationMapping -> runWithTransformation(
				transformationService.load(transformationProperties.getIdentifier(), aggregationMapping),
				createAggregator(aggregationMapping.getScope()).withAggregationMapping(aggregationMapping)));
	}

	void runWithTransformation(Transformation transformation, Aggregator aggregator) {
		try (Execution execution = new Execution(transformation)) {
			ThreadContext.put(MDC_EXECUTION, execution.id.toString());
			aggregator.withExecution(execution).aggregate();
		} finally {
			ThreadContext.remove(MDC_EXECUTION);
		}
	}

	Aggregator createAggregator(AggregationMapping.Scope scope) {
		return createAggregator()
				.withBatchSize(aggregationManagerProperties.getFetchingBatchSize())
				.withLoader(loaderRegistry.getLoader(scope));
	}

	@Lookup
	protected abstract Aggregator createAggregator();

	@Getter
	@RequiredArgsConstructor
	protected static class Execution implements AutoCloseable {
		private final UUID id = UUID.randomUUID();
		private final Transformation transformation;

		@Override
		public void close() {
			// Nothing to do here
		}
	}

}
