package de.ozgcloud.aggregation;

import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;
import java.util.function.Function;
import java.util.function.IntFunction;
import java.util.function.IntPredicate;
import java.util.function.Predicate;
import java.util.function.UnaryOperator;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import org.apache.logging.log4j.CloseableThreadContext;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import de.ozgcloud.aggregation.AggregationManagerRunner.Execution;
import de.ozgcloud.aggregation.transformation.AggregationMapping;
import de.ozgcloud.aggregation.transformation.AggregationMapping.FormIdentifier;
import de.ozgcloud.aggregation.transformation.TransformationException;
import de.ozgcloud.aggregation.transformation.VorgangMapper;
import de.ozgcloud.aggregation.warehouse.DocumentEntry;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgang;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangQuery;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangQuery.FormIdentification;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangService;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStub;
import de.ozgcloud.apilib.vorgang.Page;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Component
@Scope("prototype")
@RequiredArgsConstructor
@Log4j2
class Aggregator {

	private static final String MDC_VORGANG = "vorgang";
	private static final Predicate<Batch> HAS_NEXT_BATCH = x -> !x.items.isEmpty();

	private final OzgCloudVorgangService vorgangService;
	private final VorgangMapper vorgangMapper;

	private Execution execution;
	private FormIdentifier formIdentifier;
	private String aggregationName;
	private int batchSize = 100;
	private AggregationDataLoader loader;

	public Aggregator withExecution(Execution execution) {
		this.execution = execution;
		return this;
	}

	public Aggregator withAggregationMapping(AggregationMapping aggregationMapping) {
		if (Objects.nonNull(aggregationMapping)) {
			this.formIdentifier = aggregationMapping.getFormIdentifier();
			this.aggregationName = aggregationMapping.getName();
		}
		return this;
	}

	public Aggregator withBatchSize(int batchSize) {
		this.batchSize = batchSize;
		return this;
	}

	public Aggregator withLoader(AggregationDataLoader loader) {
		this.loader = loader;
		return this;
	}

	public void aggregate() {
		loadVorgaenge(Stream.concat(extractBatchesOfVorgaengeFromDataSource(), extractBatchesOfDeletedVorgaengeFromDataSource()));
	}

	Stream<Batch> extractBatchesOfVorgaengeFromDataSource() {
		return extractBatchesFromDataSource(this::getVorgaengeFromDataSource);
	}

	List<OzgCloudVorgang> getVorgaengeFromDataSource(Page page) {
		return vorgangService.find(buildFindByFormEngineQuery(), page).stream()
				.map(vorgangStub -> vorgangService.getById(vorgangStub.getId()))
				.toList();
	}

	OzgCloudVorgangQuery buildFindByFormEngineQuery() {
		return OzgCloudVorgangQuery.builder()
				.form(mapToFormIdentification())
				.build();
	}

	private Optional<FormIdentification> mapToFormIdentification() {
		return Optional.ofNullable(formIdentifier)
				.map(identifier -> FormIdentification.builder()
						.formId(identifier.getFormId())
						.formEngineName(identifier.getFormEngineName())
						.build());
	}

	Stream<Batch> extractBatchesOfDeletedVorgaengeFromDataSource() {
		return Objects.isNull(formIdentifier) ? extractBatchesFromDataSource(getPagedDeletedVorgaenge(vorgangService.findDeleted()))
				: Stream.empty();
	}

	Function<Page, List<OzgCloudVorgang>> getPagedDeletedVorgaenge(Stream<OzgCloudVorgangStub> allDeletedVorgaenge) {
		var it = allDeletedVorgaenge.iterator();
		IntPredicate hasNextVorgangStub = ignored -> it.hasNext();
		IntFunction<OzgCloudVorgangStub> nextVorgangStub = ignored -> it.next();
		return page -> IntStream.range(page.getOffset(), page.getOffset() + page.getLimit())
				.takeWhile(hasNextVorgangStub)
				.mapToObj(nextVorgangStub)
				.map(vorgangMapper::fromVorgangStub)
				.toList();
	}

	Stream<Batch> extractBatchesFromDataSource(Function<Page, List<OzgCloudVorgang>> getFromDataSource) {
		var initialBatch = createBatch(execution, Page.builder().offset(0).limit(batchSize).build(), getFromDataSource);
		UnaryOperator<Batch> nextBatch = x -> createBatch(
				execution,
				Page.builder().offset(x.page.getOffset() + batchSize).limit(batchSize).build(),
				getFromDataSource);
		return Stream.iterate(initialBatch, HAS_NEXT_BATCH, nextBatch).filter(x -> !x.items.isEmpty());
	}

	private Batch createBatch(Execution execution, Page page, Function<Page, List<OzgCloudVorgang>> getFromDataSource) {
		return new Batch(execution, createBatchUUID(), page, getFromDataSource.apply(page));
	}

	UUID createBatchUUID() {
		return UUID.randomUUID();
	}

	void loadVorgaenge(Stream<Batch> batches) {
		loader.loadIntoTarget(new Aggregation(aggregationName, batches.flatMap(this::transformBatchToDocumentEntries)));
	}

	Stream<DocumentEntry> transformBatchToDocumentEntries(Batch batch) {
		return batch.items.stream().map(vorgang -> transformWithinBatch(batch, vorgang)).filter(Objects::nonNull);
	}

	DocumentEntry transformWithinBatch(Batch batch, OzgCloudVorgang vorgang) {
		try (var instance = CloseableThreadContext.put(MDC_VORGANG, vorgang.getId().toString())) {
			return batch.execution.getTransformation().apply(vorgang);
		} catch (TransformationException e) {
			LOG.error(
					"failed to transform vorgang [%s] in batch [%s] of execution [%s]".formatted(vorgang.getId(), batch.id, batch.execution.getId()),
					e);
			return null;
		}
	}

	record Batch(Execution execution, UUID id, Page page, List<OzgCloudVorgang> items) {
	}

}
