/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation.extern;

import java.util.Iterator;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Future;
import java.util.stream.Stream;

import de.ozgcloud.aggregation.Aggregation;
import de.ozgcloud.aggregation.AggregationManagerProperties;
import de.ozgcloud.aggregation.data.AggregationDataServiceGrpc;
import de.ozgcloud.aggregation.data.GrpcAggregationData;
import de.ozgcloud.aggregation.data.GrpcSendAggregationDataRequest;
import de.ozgcloud.aggregation.data.GrpcSendAggregationDataResponse;
import de.ozgcloud.aggregation.warehouse.DocumentEntry;
import io.grpc.stub.ClientCallStreamObserver;
import io.grpc.stub.ClientResponseObserver;
import lombok.Builder;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import net.devh.boot.grpc.client.inject.GrpcClient;

@RequiredArgsConstructor
public class AggregationDataRemoteService {

	@GrpcClient("aggregation-manager-server")
	private final AggregationDataServiceGrpc.AggregationDataServiceStub serviceStub;
	private final AggregationManagerProperties properties;
	private final GrpcAggregationDataMapper grpcAggregationDataMapper;

	public Future<Void> sendAggregationData(Aggregation aggregation) {
		var responseObserver = buildSendAggregationDataResponseObserver(aggregation);
		serviceStub.sendAggregationData(responseObserver);
		return responseObserver.getResponseFuture();
	}

	SendAggregationDataResponseObserver buildSendAggregationDataResponseObserver(Aggregation aggregation) {
		var requestData = new RequestData(properties.getMandant(), aggregation.aggregationName(),
				toGrpcAggregationDataStream(aggregation.documentEntries()).iterator());
		return new SendAggregationDataResponseObserver(properties.getFetchingBatchSize(), requestData);
	}

	Stream<GrpcAggregationData> toGrpcAggregationDataStream(Stream<DocumentEntry> documentEntries) {
		return documentEntries.map(grpcAggregationDataMapper::toGrpcAggregationData);
	}

	@RequiredArgsConstructor
	static class SendAggregationDataResponseObserver
			implements ClientResponseObserver<GrpcSendAggregationDataRequest, GrpcSendAggregationDataResponse> {

		private final int batchSize;
		private final RequestData requestData;
		@Getter
		private final CompletableFuture<Void> responseFuture = new CompletableFuture<>();

		@Override
		public void beforeStart(ClientCallStreamObserver<GrpcSendAggregationDataRequest> requestObserver) {
			requestObserver.setOnReadyHandler(buildOnReadyHandler(requestObserver));
		}

		@Override
		public void onNext(GrpcSendAggregationDataResponse value) {

		}

		@Override
		public void onError(Throwable t) {
			responseFuture.completeExceptionally(t);
		}

		@Override
		public void onCompleted() {
			responseFuture.complete(null);
		}

		SendAggregationDataOnReadyHandler buildOnReadyHandler(ClientCallStreamObserver<GrpcSendAggregationDataRequest> requestObserver) {
			return new SendAggregationDataOnReadyHandler(requestObserver, batchSize, requestData);
		}
	}

	@RequiredArgsConstructor
	static class SendAggregationDataOnReadyHandler implements Runnable {

		private final ClientCallStreamObserver<GrpcSendAggregationDataRequest> requestObserver;
		private final int batchSize;
		private final RequestData requestData;

		@Override
		public void run() {
			while (requestObserver.isReady() && requestData.aggregationDataIterator.hasNext()) {
				requestObserver.onNext(buildRequest());
			}
			if (!requestData.aggregationDataIterator.hasNext()) {
				requestObserver.onCompleted();
			}
		}

		GrpcSendAggregationDataRequest buildRequest() {
			var builder = GrpcSendAggregationDataRequest.newBuilder()
					.setName(requestData.aggregationName())
					.setMandant(requestData.mandant());
			addAggregationData(builder);
			return builder.build();
		}

		private void addAggregationData(GrpcSendAggregationDataRequest.Builder builder) {
			var elementsAdded = 0;
			while (requestData.aggregationDataIterator.hasNext() && elementsAdded++ < batchSize) {
				builder.addAggregationData(requestData.aggregationDataIterator.next());
			}
		}
	}

	@Builder
	record RequestData(String mandant, String aggregationName, Iterator<GrpcAggregationData> aggregationDataIterator) {
	}
}
