/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.MockedStatic;

import de.ozgcloud.aggregation.transformation.AggregationMapping;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.common.test.ReflectionTestUtils;
import lombok.Getter;
import lombok.RequiredArgsConstructor;

class AggregationDataLoaderRegistryTest {

	private final AggregationMapping.Scope registeredScope = AggregationMapping.Scope.INTERN;
	private final AggregationMapping.Scope notRegisteredScope = AggregationMapping.Scope.EXTERN;

	private final AggregationDataLoader loader = new TestLoader(registeredScope);
	private AggregationDataLoaderRegistry registry;

	@BeforeEach
	void init() {
		registry = spy(new AggregationDataLoaderRegistry(List.of(loader)));
	}

	@Nested
	class TestConstructor {

		private final List<AggregationDataLoader> loaders = List.of(new TestLoader(AggregationMapping.Scope.INTERN));

		@Test
		void shouldConvertToLoadersByScope() {
			try (MockedStatic<AggregationDataLoaderRegistry> mocked = mockStatic(AggregationDataLoaderRegistry.class)) {
				new AggregationDataLoaderRegistry(loaders);
				mocked.verify(() -> AggregationDataLoaderRegistry.toLoadersByScope(loaders));
			}
		}

		@Test
		@SuppressWarnings("unchecked")
		void setLoadersByScope() {
			var loadersByScope = Map.of(AggregationMapping.Scope.INTERN, new TestLoader(AggregationMapping.Scope.INTERN));
			try (MockedStatic<AggregationDataLoaderRegistry> mocked = mockStatic(AggregationDataLoaderRegistry.class)) {
				mocked.when(() -> AggregationDataLoaderRegistry.toLoadersByScope(loaders)).thenReturn(loadersByScope);

				AggregationDataLoaderRegistry registry = new AggregationDataLoaderRegistry(loaders);

				var loadersByScopeFieldValue = ReflectionTestUtils.getField(registry, "loadersByScope", Map.class);
				assertThat(loadersByScopeFieldValue).isEqualTo(loadersByScope);
			}
		}
	}

	@Nested
	class TestHasLoader {

		@Test
		void shouldReturnFalseIfNoLoaderRegisteredForScope() {
			var hasLoader = registry.hasLoader(notRegisteredScope);

			assertThat(hasLoader).isFalse();
		}

		@Test
		void shouldReturnTrueIfLoaderRegisteredForScope() {
			var hasLoader = registry.hasLoader(registeredScope);

			assertThat(hasLoader).isTrue();
		}
	}

	@Nested
	class TestGetLoader {

		@Test
		void shouldReturnLoaderThatSupportsScope() {
			var loaderSupportingScope = registry.getLoader(registeredScope);

			assertThat(loaderSupportingScope).isSameAs(loader);
		}

		@Test
		void shouldThrowExceptionIfNoLoaderRegisteredForScope() {
			assertThatThrownBy(() -> registry.getLoader(notRegisteredScope)).isInstanceOf(TechnicalException.class);
		}
	}

	@Nested
	class TestToLoadersByScope {

		@ParameterizedTest
		@EnumSource
		void shouldAddLoader(AggregationMapping.Scope scope) {
			var loader = new TestLoader(scope);
			var loadersByScope = AggregationDataLoaderRegistry.toLoadersByScope(List.of(loader));

			assertThat(loadersByScope).containsEntry(scope, loader);
		}

		@Test
		void shouldThrowExceptionIfMultipleLoadersHaveSameScope() {
			var scope = AggregationMapping.Scope.EXTERN;
			assertThatThrownBy(() -> AggregationDataLoaderRegistry.toLoadersByScope(List.of(new TestLoader(scope), new TestLoader(scope))))
					.isInstanceOf(TechnicalException.class);
		}
	}

	@RequiredArgsConstructor
	static class TestLoader implements AggregationDataLoader {

		@Getter
		private final AggregationMapping.Scope scope;

		@Override
		public void loadIntoTarget(Aggregation aggregation) {

		}
	}
}
