/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.stream.Stream;

import org.apache.commons.lang3.RandomUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.AdditionalMatchers;
import org.mockito.ArgumentMatcher;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.aggregation.AggregationManagerRunner.Execution;
import de.ozgcloud.aggregation.transformation.AggregationMapping;
import de.ozgcloud.aggregation.transformation.AggregationMappingTestFactory;
import de.ozgcloud.aggregation.transformation.Transformation;
import de.ozgcloud.aggregation.transformation.TransformationService;

class AggregationManagerRunnerTest {

	@Mock
	private AggregationManagerProperties aggregationManagerProperties;
	@Mock
	private TransformationProperties transformationProperties;
	@Mock
	private TransformationService transformationService;
	@Mock
	private AggregationDataLoaderRegistry loaderRegistry;
	@Mock
	private static Aggregator aggregator;
	@Spy
	@InjectMocks
	private AggregationManagerRunnerImpl runner;

	@Nested
	class TestRun {

		@BeforeEach
		void init() {
			doNothing().when(runner).runWithConfiguredTransformations();
		}

		@Test
		void shouldGetScopesWithoutConfiguredTransformations() {
			runner.run();

			verify(runner).getScopesWithoutConfiguredTransformations();
		}

		@ParameterizedTest
		@EnumSource
		void shouldRunWithDefaultTransformationForScopesWithoutConfiguredOnes(AggregationMapping.Scope scope) {
			doReturn(Stream.of(scope)).when(runner).getScopesWithoutConfiguredTransformations();
			doNothing().when(runner).runWithDefaultTransformation(any());

			runner.run();

			verify(runner).runWithDefaultTransformation(scope);
		}

		@Test
		void shouldRunWithConfiguredTransformations() {
			runner.run();

			verify(runner).runWithConfiguredTransformations();
		}
	}

	@Nested
	class TestGetScopesWithoutConfiguredTransformations {

		@ParameterizedTest
		@EnumSource
		void shouldCheckForLoader(AggregationMapping.Scope scope) {
			runner.getScopesWithoutConfiguredTransformations().toList();

			verify(loaderRegistry).hasLoader(scope);
		}

		@ParameterizedTest
		@EnumSource
		void shouldCheckForConfiguredTransformationsWithScope(AggregationMapping.Scope scope) {
			when(loaderRegistry.hasLoader(any())).thenReturn(true);

			runner.getScopesWithoutConfiguredTransformations().toList();

			verify(runner).hasNoConfiguredTransformations(scope);
		}

		@ParameterizedTest
		@EnumSource
		void shouldReturnScopesWithoutTransformations(AggregationMapping.Scope scope) {
			doReturn(true).when(runner).hasNoConfiguredTransformations(scope);
			doReturn(false).when(runner).hasNoConfiguredTransformations(AdditionalMatchers.not(eq(scope)));
			when(loaderRegistry.hasLoader(any())).thenReturn(true);

			var scopes = runner.getScopesWithoutConfiguredTransformations();

			assertThat(scopes).containsExactly(scope);
		}

		@Test
		void shouldOmitScopesWithoutLoaders() {
			lenient().doReturn(true).when(runner).hasNoConfiguredTransformations(any());
			when(loaderRegistry.hasLoader(any())).thenReturn(false);

			var scopes = runner.getScopesWithoutConfiguredTransformations();

			assertThat(scopes).isEmpty();
		}
	}

	@Nested
	class TestHasNoConfiguredTransformations {

		@Test
		void shouldGetAggregationMappings() {
			runner.hasNoConfiguredTransformations(AggregationMapping.Scope.INTERN);

			verify(transformationProperties).getAggregationMappings();
		}

		@Test
		void shouldReturnFalseIfThereAreConfiguredTransformations() {
			when(transformationProperties.getAggregationMappings()).thenReturn(List.of(AggregationMappingTestFactory.create()));

			var hasNoTransformations = runner.hasNoConfiguredTransformations(AggregationMappingTestFactory.SCOPE);

			assertThat(hasNoTransformations).isFalse();
		}

		@Test
		void shouldReturnTrueIfThereAreNoConfiguredTransformations() {
			when(transformationProperties.getAggregationMappings()).thenReturn(List.of(AggregationMappingTestFactory.createBuilder().scope(
					AggregationMapping.Scope.INTERN).build()));

			var hasNoTransformations = runner.hasNoConfiguredTransformations(AggregationMapping.Scope.EXTERN);

			assertThat(hasNoTransformations).isTrue();
		}
	}

	@Nested
	class TestRunWithDefaultTransformation {

		private final AggregationMapping.Scope scope = AggregationMapping.Scope.INTERN;
		private final String identifier = LoremIpsum.getInstance().getWords(1);

		@Mock
		private Transformation transformation;

		@BeforeEach
		void init() {
			when(transformationProperties.getIdentifier()).thenReturn(identifier);
			when(transformationService.load(any(), any())).thenReturn(transformation);
			doReturn(aggregator).when(runner).createAggregator(any());
			doNothing().when(runner).runWithTransformation(any(), any());
		}

		@Test
		void shouldGetIdentifier() {
			runner.runWithDefaultTransformation(scope);

			verify(transformationProperties).getIdentifier();
		}

		@Test
		void shouldLoadTransformation() {
			runner.runWithDefaultTransformation(scope);

			verify(transformationService).load(identifier, null);
		}

		@Test
		void shouldCreateAggregatorForScope() {
			runner.runWithDefaultTransformation(scope);

			verify(runner).createAggregator(scope);
		}

		@Test
		void shouldRunWithTransformation() {
			runner.runWithDefaultTransformation(scope);

			verify(runner).runWithTransformation(transformation, aggregator);
		}
	}

	@Nested
	class TestRunWithConfiguredTransformations {

		private final AggregationMapping.Scope firstScope = AggregationMapping.Scope.INTERN;
		@Mock
		private Transformation firstTransformation;
		@Mock
		private Aggregator firstAggregator;

		private final AggregationMapping.Scope secondScope = AggregationMapping.Scope.EXTERN;
		@Mock
		private Transformation secondTransformation;
		@Mock
		private Aggregator secondAggregator;

		private final String identifier = LoremIpsum.getInstance().getWords(1);
		private final List<AggregationMapping> aggregationMappings = List.of(
				AggregationMappingTestFactory.createBuilder().scope(firstScope).build(),
				AggregationMappingTestFactory.createBuilder().scope(secondScope).build());

		@BeforeEach
		void mock() {
			when(transformationProperties.getAggregationMappings()).thenReturn(aggregationMappings);
			when(transformationProperties.getIdentifier()).thenReturn(identifier);
			when(transformationService.load(any(), any())).thenReturn(firstTransformation, secondTransformation);
			doNothing().when(runner).runWithTransformation(any(), any());
			doReturn(firstAggregator).when(runner).createAggregator(firstScope);
			doReturn(secondAggregator).when(runner).createAggregator(secondScope);
			when(firstAggregator.withAggregationMapping(any())).thenReturn(firstAggregator);
			when(secondAggregator.withAggregationMapping(any())).thenReturn(secondAggregator);
		}

		@Test
		void shouldGetIdentifier() {
			runner.runWithConfiguredTransformations();

			verify(transformationProperties, times(2)).getIdentifier();
		}

		@Test
		void shouldLoadTransformationForEachAggregationMapping() {
			runner.runWithConfiguredTransformations();

			aggregationMappings.forEach(mapping -> verify(transformationService).load(identifier, mapping));
		}

		@Test
		void shouldCreateAggregatorForScope() {
			runner.runWithConfiguredTransformations();

			verify(runner).createAggregator(firstScope);
			verify(runner).createAggregator(secondScope);
		}

		@Test
		void shouldSetAggregationMappingInAggregator() {
			runner.runWithConfiguredTransformations();

			verify(firstAggregator).withAggregationMapping(aggregationMappings.getFirst());
			verify(secondAggregator).withAggregationMapping(aggregationMappings.getLast());
		}

		@Test
		void shouldRunWithTransformationForEachTransformation() {
			runner.runWithConfiguredTransformations();

			verify(runner).runWithTransformation(firstTransformation, firstAggregator);
			verify(runner).runWithTransformation(secondTransformation, secondAggregator);
		}
	}

	@Nested
	class TestRunWithTransformation {

		@Mock
		private Transformation transformation;
		private final ArgumentMatcher<Execution> hasTransformation = execution -> execution.getTransformation().equals(transformation);

		@BeforeEach
		void init() {
			when(aggregator.withExecution(any())).thenReturn(aggregator);
		}

		@Test
		void shouldSetExecution() {
			runWithTransformation();

			verify(aggregator).withExecution(argThat(hasTransformation));
		}

		@Test
		void shouldLoadAggregation() {
			runWithTransformation();

			verify(aggregator).aggregate();
		}

		private void runWithTransformation() {
			runner.runWithTransformation(transformation, aggregator);
		}
	}

	@Nested
	class TestCreateAggregator {

		private final AggregationMapping.Scope scope = AggregationMapping.Scope.INTERN;
		private final int batchSize = RandomUtils.insecure().randomInt();
		@Mock
		private AggregationDataLoader loader;

		@BeforeEach
		void init() {
			when(aggregationManagerProperties.getFetchingBatchSize()).thenReturn(batchSize);
			when(loaderRegistry.getLoader(any())).thenReturn(loader);
			when(aggregator.withBatchSize(anyInt())).thenReturn(aggregator);
			when(aggregator.withLoader(any())).thenReturn(aggregator);
		}

		@Test
		void shouldGetBatchSize() {
			runner.createAggregator(scope);

			verify(aggregationManagerProperties).getFetchingBatchSize();
		}

		@Test
		void shouldSetBatchSize() {
			runner.createAggregator(scope);

			verify(aggregator).withBatchSize(batchSize);
		}

		@Test
		void shouldGetLoaderFromRegistry() {
			runner.createAggregator(scope);

			verify(loaderRegistry).getLoader(scope);
		}

		@Test
		void shouldSetLoader() {
			runner.createAggregator(scope);

			verify(aggregator).withLoader(loader);
		}

		@Test
		void shouldReturnAggregator() {
			var created = runner.createAggregator(scope);

			assertThat(created).isEqualTo(aggregator);
		}
	}

	static class AggregationManagerRunnerImpl extends AggregationManagerRunner {

		public AggregationManagerRunnerImpl(AggregationManagerProperties aggregationManagerProperties, TransformationProperties transformationProperties,
				TransformationService transformationService, AggregationDataLoaderRegistry loaderRegistry) {
			super(aggregationManagerProperties, transformationProperties, transformationService, loaderRegistry);
		}

		@Override
		protected Aggregator createAggregator() {
			return aggregator;
		}
	}
}
