package de.ozgcloud.aggregation;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Stream;

import org.apache.commons.lang3.RandomUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.ArgumentMatcher;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.aggregation.AggregationManagerRunner.Execution;
import de.ozgcloud.aggregation.Aggregator.Batch;
import de.ozgcloud.aggregation.transformation.AggregationMapping;
import de.ozgcloud.aggregation.transformation.AggregationMapping.FormIdentifier;
import de.ozgcloud.aggregation.transformation.AggregationMappingTestFactory;
import de.ozgcloud.aggregation.transformation.FormIdentifierTestFactory;
import de.ozgcloud.aggregation.transformation.Transformation;
import de.ozgcloud.aggregation.transformation.VorgangMapper;
import de.ozgcloud.aggregation.warehouse.DocumentEntry;
import de.ozgcloud.aggregation.warehouse.DocumentEntryTestFactory;
import de.ozgcloud.aggregation.warehouse.WarehouseRepository;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgang;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangQuery;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangService;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStub;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangStubTestFactory;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangTestFactory;
import de.ozgcloud.apilib.vorgang.Page;

class AggregatorTest {

	@InjectMocks
	@Spy
	private Aggregator aggregator;
	@Mock
	private Transformation transformation;
	private final Execution execution = new Execution(transformation);
	@Mock
	private WarehouseRepository repository;
	@Mock
	private VorgangMapper vorgangMapper;
	@Mock
	private OzgCloudVorgangService vorgangService;

	@Nested
	class TestWithExecution {

		@Test
		void shouldReturnSelf() {
			var result = aggregator.withExecution(execution);

			assertThat(result).isSameAs(aggregator);
		}

		@Test
		void shouldSetExecution() {
			var result = aggregator.withExecution(execution);

			assertThat(ReflectionTestUtils.getField(result, "execution")).isEqualTo(execution);
		}
	}

	@Nested
	class TestWithAggregationMapping {

		@Nested
		class TestOnNonNullAggregationMapping {

			private final AggregationMapping aggregationMapping = AggregationMappingTestFactory.create();
			private final FormIdentifier formIdentifier = AggregationMappingTestFactory.FORM_IDENTIFIER;
			private final String aggregationName = AggregationMappingTestFactory.NAME;

			@Test
			void shouldReturnSelf() {
				var result = aggregator.withAggregationMapping(aggregationMapping);

				assertThat(result).isSameAs(aggregator);
			}

			@Test
			void shouldSetFormIdentifier() {
				var result = aggregator.withAggregationMapping(aggregationMapping);

				assertThat(ReflectionTestUtils.getField(result, "formIdentifier")).isEqualTo(formIdentifier);
			}

			@Test
			void shouldSetAggregationName() {
				var result = aggregator.withAggregationMapping(aggregationMapping);

				assertThat(ReflectionTestUtils.getField(result, "aggregationName")).isEqualTo(aggregationName);
			}
		}

		@Nested
		class TestOnNullAggregationMapping {

			@Test
			void shouldReturnSelf() {
				var result = aggregator.withAggregationMapping(null);

				assertThat(result).isSameAs(aggregator);
			}

			@Test
			void shouldNotSetFormIdentifier() {
				var result = aggregator.withAggregationMapping(null);

				assertThat(ReflectionTestUtils.getField(result, "formIdentifier")).isNull();
			}

			@Test
			void shouldNotSetAggregationName() {
				var result = aggregator.withAggregationMapping(null);

				assertThat(ReflectionTestUtils.getField(result, "aggregationName")).isNull();
			}
		}
	}

	@Nested
	class TestWithBatchSize {

		private final int batchSize = RandomUtils.insecure().randomInt();

		@Test
		void shouldReturnSelf() {
			var result = aggregator.withBatchSize(batchSize);

			assertThat(result).isSameAs(aggregator);
		}

		@Test
		void shouldSetBatchSize() {
			var result = aggregator.withBatchSize(batchSize);

			assertThat(ReflectionTestUtils.getField(result, "batchSize")).isEqualTo(batchSize);
		}
	}

	@Nested
	class TestWithLoader {

		@Mock
		private AggregationDataLoader loader;

		@Test
		void shouldReturnSelf() {
			var result = aggregator.withLoader(loader);

			assertThat(result).isSameAs(aggregator);
		}

		@Test
		void shouldSetLoader() {
			aggregator.withLoader(loader);

			assertThat(ReflectionTestUtils.getField(aggregator, "loader")).isSameAs(loader);
		}
	}

	@Nested
	class TestAggregate {
		@Mock
		private Batch batchOfVorgaenge;
		@Mock
		private Batch batchOfDeletedVorgaenge;
		@Captor
		private ArgumentCaptor<Stream<Batch>> batchStreamCaptor;

		@BeforeEach
		void setUp() {
			doReturn(Stream.of(batchOfVorgaenge)).when(aggregator).extractBatchesOfVorgaengeFromDataSource();
			doReturn(Stream.of(batchOfDeletedVorgaenge)).when(aggregator).extractBatchesOfDeletedVorgaengeFromDataSource();
			doNothing().when(aggregator).loadVorgaenge(any());
		}

		@Test
		void shouldExtractBatchesOfVorgaengeFromDataSource() {
			aggregator.aggregate();

			verify(aggregator).extractBatchesOfVorgaengeFromDataSource();
		}

		@Test
		void shouldExtractBatchesOfDeletedVorgaengeFromDataSource() {
			aggregator.aggregate();

			verify(aggregator).extractBatchesOfDeletedVorgaengeFromDataSource();
		}

		@Test
		void shouldLoadVorgaengeIntoRepository() {
			aggregator.aggregate();

			verify(aggregator).loadVorgaenge(batchStreamCaptor.capture());
			assertThat(batchStreamCaptor.getValue()).containsExactly(batchOfVorgaenge, batchOfDeletedVorgaenge);
		}
	}

	@Nested
	class TestExtractBatchesOfVorgaengeFromDataSource {
		@Captor
		private ArgumentCaptor<Function<Page, List<OzgCloudVorgang>>> functionToRetrieveDataCaptor;
		@Mock
		private Batch batch;
		@Mock
		private Page page;

		@BeforeEach
		void init() {
			doReturn(Stream.of(batch)).when(aggregator).extractBatchesFromDataSource(any());
		}

		@Test
		void shouldExtractWithDataRetrievalFunction() {
			doReturn(Collections.emptyList()).when(aggregator).getVorgaengeFromDataSource(any());

			aggregator.extractBatchesOfVorgaengeFromDataSource();

			verify(aggregator).extractBatchesFromDataSource(functionToRetrieveDataCaptor.capture());
			functionToRetrieveDataCaptor.getValue().apply(page);
			verify(aggregator).getVorgaengeFromDataSource(page);
		}

		@Test
		void shouldReturnExtractedBatches() {
			var extracted = aggregator.extractBatchesOfVorgaengeFromDataSource();

			assertThat(extracted).containsExactly(batch);
		}
	}

	@Nested
	class TestGetVorgaengeFromDataSource {

		private final Page page = Page.builder().offset(10).limit(2).build();
		private final OzgCloudVorgangQuery query = OzgCloudVorgangQuery.builder().build();

		@BeforeEach
		void init() {
			doReturn(query).when(aggregator).buildFindByFormEngineQuery();
			when(vorgangService.find(any(), any())).thenReturn(List.of(OzgCloudVorgangStubTestFactory.create()));
			when(vorgangService.getById(any())).thenReturn(OzgCloudVorgangTestFactory.create());
		}

		@Test
		void shouldCallBuildFindByFormEngineQuery() {
			getVorgaengeFromDataSource();

			verify(aggregator).buildFindByFormEngineQuery();
		}

		@Test
		void shouldCallVorgangService() {
			getVorgaengeFromDataSource();

			verify(vorgangService).find(query, page);
		}

		@Test
		void shouldGetVorgangDetails() {
			getVorgaengeFromDataSource();

			verify(vorgangService).getById(OzgCloudVorgangTestFactory.ID);
		}

		@Test
		void shouldReturnVorgangDetails() {
			var vorgaenge = getVorgaengeFromDataSource();

			assertThat(vorgaenge).usingRecursiveFieldByFieldElementComparator().containsExactly(OzgCloudVorgangTestFactory.create());
		}

		private List<OzgCloudVorgang> getVorgaengeFromDataSource() {
			return aggregator.getVorgaengeFromDataSource(page);
		}
	}

	@Nested
	class TestBuildFindByFormEngineQuery {

		@Nested
		class TestOnFormIdentifierNull {

			@Test
			void shouldReturnFindAllQueryOnNullFormIdentifier() {
				var query = aggregator.buildFindByFormEngineQuery();

				assertThat(query).usingRecursiveComparison().isEqualTo(OzgCloudVorgangQuery.builder().build());
			}
		}

		@Nested
		class TestOnFormIdentifierNotNull {

			private final AggregationMapping aggregationMapping = AggregationMappingTestFactory.create();

			@BeforeEach
			void init() {
				aggregator = aggregator.withAggregationMapping(aggregationMapping);
			}

			@Test
			void shouldSetFormIdInQuery() {
				var query = aggregator.buildFindByFormEngineQuery();

				assertThat(query.getForm()).get().extracting("formId").isEqualTo(FormIdentifierTestFactory.FORM_ID);
			}

			@Test
			void shouldSetFormEngineNameInQuery() {
				var query = aggregator.buildFindByFormEngineQuery();

				assertThat(query.getForm()).get().extracting("formEngineName").isEqualTo(FormIdentifierTestFactory.FORM_ENGINE_NAME);
			}
		}
	}

	@Nested
	class TestExtractBatchesOfDeletedVorgaengeFromDataSource {

		@Mock
		private Execution execution;
		@Mock
		private Function<Page, List<OzgCloudVorgang>> functionToRetrieveData;
		private final List<OzgCloudVorgangStub> deletedVorgaenge = List.of(OzgCloudVorgangStubTestFactory.create());
		@Captor
		private ArgumentCaptor<Stream<OzgCloudVorgangStub>> deletedVorgaengeCaptor;
		@Mock
		private Batch batch;

		@Nested
		class TestOnFormIdentifierNull {

			@BeforeEach
			void init() {
				when(vorgangService.findDeleted()).thenReturn(deletedVorgaenge.stream());
				doReturn(functionToRetrieveData).when(aggregator).getPagedDeletedVorgaenge(any());
				doReturn(Stream.of(batch)).when(aggregator).extractBatchesFromDataSource(any());
			}

			@Test
			void shouldFindDeleted() {
				aggregator.extractBatchesOfDeletedVorgaengeFromDataSource();

				verify(vorgangService).findDeleted();
			}

			@Test
			void shouldGetPagedDeletedVorgaenge() {
				aggregator.extractBatchesOfDeletedVorgaengeFromDataSource();

				verify(aggregator).getPagedDeletedVorgaenge(deletedVorgaengeCaptor.capture());
				assertThat(deletedVorgaengeCaptor.getValue()).usingRecursiveFieldByFieldElementComparator()
						.containsExactlyElementsOf(deletedVorgaenge);
			}

			@Test
			void shouldExtractWithDataRetrievalFunction() {
				aggregator.extractBatchesOfDeletedVorgaengeFromDataSource();

				verify(aggregator).extractBatchesFromDataSource(functionToRetrieveData);
			}

			@Test
			void shouldReturnExtractedBatches() {
				var extracted = aggregator.extractBatchesOfDeletedVorgaengeFromDataSource();

				assertThat(extracted).containsExactly(batch);
			}
		}

		@Nested
		class TestOnFormIdentifierNotNull {

			private final AggregationMapping aggregationMapping = AggregationMappingTestFactory.create();

			@BeforeEach
			void init() {
				aggregator = aggregator.withAggregationMapping(aggregationMapping);
			}

			@Test
			void shouldNotFindDeleted() {
				aggregator.extractBatchesOfDeletedVorgaengeFromDataSource();

				verify(vorgangService, never()).findDeleted();
			}

			@Test
			void shouldReturnEmptyStream() {
				var extracted = aggregator.extractBatchesOfDeletedVorgaengeFromDataSource();

				assertThat(extracted).isEmpty();
			}
		}
	}

	@Nested
	class TestGetPagedDeletedVorgaenge {

		private static final int PAGE_SIZE = 2;

		private final Stream<OzgCloudVorgangStub> vorgangStubs = Stream.generate(OzgCloudVorgangStubTestFactory::create).limit(PAGE_SIZE + 1);

		@Test
		void shouldReturnFirstFullPage() {
			var pagingFunction = aggregator.getPagedDeletedVorgaenge(vorgangStubs);

			var vorgaenge = getFirstPage(pagingFunction);

			assertThat(vorgaenge).hasSize(PAGE_SIZE);
		}

		@Test
		void shouldReturnSecondPageWithOneElement() {
			var pagingFunction = aggregator.getPagedDeletedVorgaenge(vorgangStubs);
			getFirstPage(pagingFunction);

			var vorgaenge = getSecondPage(pagingFunction);

			assertThat(vorgaenge).hasSize(1);
		}

		private List<OzgCloudVorgang> getFirstPage(Function<Page, List<OzgCloudVorgang>> pagingFunction) {
			return pagingFunction.apply(Page.builder().offset(0).limit(PAGE_SIZE).build());
		}

		private List<OzgCloudVorgang> getSecondPage(Function<Page, List<OzgCloudVorgang>> pagingFunction) {
			return pagingFunction.apply(Page.builder().offset(PAGE_SIZE).limit(PAGE_SIZE).build());
		}
	}

	@Nested
	class TestExtractBatchesFromDataSource {

		private static final int BATCH_SIZE = 2;
		private static final List<OzgCloudVorgang> VORGAENGE_1 = List.of(OzgCloudVorgangTestFactory.create(), OzgCloudVorgangTestFactory.create());
		private static final List<OzgCloudVorgang> VORGAENGE_2 = List.of(OzgCloudVorgangTestFactory.create());
		private static final UUID BATCH_1_UUID = UUID.randomUUID();
		private static final UUID BATCH_2_UUID = UUID.randomUUID();

		@Mock
		private Function<Page, List<OzgCloudVorgang>> functionToRetrieveData;

		@SuppressWarnings("unchecked")
		@BeforeEach
		void init() {
			aggregator = aggregator.withBatchSize(BATCH_SIZE).withExecution(execution);
			when(functionToRetrieveData.apply(any())).thenReturn(VORGAENGE_1, VORGAENGE_2, Collections.emptyList());
			when(aggregator.createBatchUUID()).thenReturn(BATCH_1_UUID, BATCH_2_UUID);
		}

		@Test
		void shouldCallFunctionToRetrieveDataForFirstPage() {
			extractBatchesFromDataSource();

			verify(functionToRetrieveData).apply(argThat(hasExpectedOffsetAndConfiguredLimit(0)));
		}

		@Test
		void shouldCallFunctionToRetrieveDataForSecondPage() {
			extractBatchesFromDataSource().toList();

			verify(functionToRetrieveData).apply(argThat(hasExpectedOffsetAndConfiguredLimit(BATCH_SIZE)));
		}

		@Test
		void shouldCallFunctionToRetrieveDataForThirdPage() {
			extractBatchesFromDataSource().toList();

			verify(functionToRetrieveData).apply(argThat(hasExpectedOffsetAndConfiguredLimit(BATCH_SIZE * 2)));
		}

		@Test
		void shouldReturnBatches() {
			var batches = extractBatchesFromDataSource().toList();

			assertThat(batches).hasSize(2).usingRecursiveFieldByFieldElementComparator().containsExactly(
					new Batch(execution, BATCH_1_UUID, Page.builder().offset(0).limit(BATCH_SIZE).build(), VORGAENGE_1),
					new Batch(execution, BATCH_2_UUID, Page.builder().offset(BATCH_SIZE).limit(BATCH_SIZE).build(), VORGAENGE_2));
		}

		private ArgumentMatcher<Page> hasExpectedOffsetAndConfiguredLimit(int expectedOffset) {
			return page -> page.getOffset() == expectedOffset && page.getLimit() == BATCH_SIZE;
		}

		private Stream<Batch> extractBatchesFromDataSource() {
			return aggregator.extractBatchesFromDataSource(functionToRetrieveData);
		}
	}

	@Nested
	class TestLoadVorgaenge {

		@Mock
		private Execution execution;
		@Mock
		private Batch batch;
		@Mock
		private AggregationDataLoader loader;
		private final DocumentEntry documentEntry = DocumentEntryTestFactory.create();
		private final AggregationMapping aggregationMapping = AggregationMappingTestFactory.create();
		@Captor
		private ArgumentCaptor<Aggregation> aggregationCaptor;

		@BeforeEach
		void init() {
			aggregator = aggregator.withAggregationMapping(aggregationMapping).withLoader(loader);
		}

		@Test
		void shouldLoadIntoTarget() {
			loadVorgaenge();

			verify(loader).loadIntoTarget(any(Aggregation.class));
		}

		@Test
		void shouldTransform() {
			doReturn(Stream.of(documentEntry)).when(aggregator).transformBatchToDocumentEntries(any());

			loadVorgaenge();

			getArgumentOfLoadIntoTarget().documentEntries().toList();
			verify(aggregator).transformBatchToDocumentEntries(batch);
		}

		@Test
		void shouldSetAggregationName() {
			loadVorgaenge();

			assertThat(getArgumentOfLoadIntoTarget().aggregationName()).isEqualTo(AggregationMappingTestFactory.NAME);
		}

		private Aggregation getArgumentOfLoadIntoTarget() {
			verify(loader).loadIntoTarget(aggregationCaptor.capture());
			return aggregationCaptor.getValue();
		}

		private void loadVorgaenge() {
			aggregator.loadVorgaenge(Stream.of(batch));
		}
	}
}
