/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation.extern;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.aggregation.Aggregation;
import de.ozgcloud.aggregation.AggregationTestFactory;
import de.ozgcloud.aggregation.transformation.AggregationMapping;
import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.SneakyThrows;

class AggregationDataRemoteLoaderTest {

	@Mock
	private AggregationDataRemoteService remoteService;
	@Spy
	@InjectMocks
	private AggregationDataRemoteLoader loader;

	@Nested
	class TestLoadIntoTarget {

		private final Aggregation aggregation = AggregationTestFactory.create();
		@Mock
		private Future<Void> future;

		@BeforeEach
		void init() {
			when(remoteService.sendAggregationData(any())).thenReturn(future);
		}

		@Test
		@SneakyThrows
		void shouldSendAggregationData() {
			loader.loadIntoTarget(aggregation);

			verify(remoteService).sendAggregationData(aggregation);
		}

		@Test
		@SneakyThrows
		void shouldGetFromFuture() {
			loader.loadIntoTarget(aggregation);

			verify(future).get();
		}

		@Nested
		class TestOnInterruptedException {

			private final InterruptedException exception = new InterruptedException();

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(future.get()).thenThrow(exception);
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, () -> loader.loadIntoTarget(aggregation));
			}

			@Test
			void shouldInterruptThread() {
				try {
					loader.loadIntoTarget(aggregation);
				} catch (TechnicalException e) {
					// expected
				}

				assertThat(Thread.currentThread().isInterrupted()).isTrue();
			}
		}

		@Nested
		class TestOnExecutionException {

			private final ExecutionException exception = new ExecutionException(new Exception());

			@BeforeEach
			@SneakyThrows
			void mock() {
				when(future.get()).thenThrow(exception);
			}

			@Test
			void shouldThrowTechnicalException() {
				assertThrows(TechnicalException.class, () -> loader.loadIntoTarget(aggregation));
			}
		}
	}

	@Nested
	class TestGetScope {

		@Test
		void shouldReturnExternScope() {
			var scope = loader.getScope();

			assertThat(scope).isEqualTo(AggregationMapping.Scope.EXTERN);
		}
	}
}
