/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation.extern;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Random;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Future;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import org.apache.commons.collections.IteratorUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.aggregation.Aggregation;
import de.ozgcloud.aggregation.AggregationManagerProperties;
import de.ozgcloud.aggregation.AggregationTestFactory;
import de.ozgcloud.aggregation.data.AggregationDataServiceGrpc;
import de.ozgcloud.aggregation.data.GrpcAggregationData;
import de.ozgcloud.aggregation.data.GrpcSendAggregationDataRequest;
import de.ozgcloud.aggregation.extern.AggregationDataRemoteService.SendAggregationDataResponseObserver;
import de.ozgcloud.aggregation.warehouse.DocumentEntry;
import de.ozgcloud.common.test.ReflectionTestUtils;
import io.grpc.stub.ClientCallStreamObserver;
import lombok.SneakyThrows;

class AggregationDataRemoteServiceTest {

	@Mock
	private AggregationDataServiceGrpc.AggregationDataServiceStub serviceStub;
	@Mock
	private AggregationManagerProperties properties;
	@Mock
	private GrpcAggregationDataMapper grpcAggregationDataMapper;
	@InjectMocks
	@Spy
	private AggregationDataRemoteService service;

	@Nested
	class TestSendAggregationData {

		private final Aggregation aggregation = AggregationTestFactory.create();

		@Mock
		private SendAggregationDataResponseObserver responseObserver;
		@Mock
		private CompletableFuture<Void> responseFuture;

		@BeforeEach
		void init() {
			doReturn(responseObserver).when(service).buildSendAggregationDataResponseObserver(any());
			when(responseObserver.getResponseFuture()).thenReturn(responseFuture);
		}

		@Test
		void shouldBuildResponseObserver() {
			sendAggregationData();

			verify(service).buildSendAggregationDataResponseObserver(aggregation);
		}

		@Test
		void shouldCallServiceStub() {
			sendAggregationData();

			verify(serviceStub).sendAggregationData(responseObserver);
		}

		@Test
		void shouldReturnFutureFromResponseObserver() {
			var future = sendAggregationData();

			assertThat(future).isSameAs(responseFuture);
		}

		private Future<Void> sendAggregationData() {
			return service.sendAggregationData(aggregation);
		}
	}

	@Nested
	class TestBuildSendAggregationDataResponseObserver {

		private static final int BATCH_SIZE = new Random().nextInt(100);
		@Captor
		private ArgumentCaptor<Stream<DocumentEntry>> documentEntriesCaptor;

		@BeforeEach
		void init() {
			doReturn(Stream.of(GrpcAggregationDataTestFactory.create())).when(service).toGrpcAggregationDataStream(any());
			when(properties.getFetchingBatchSize()).thenReturn(BATCH_SIZE);
		}

		@Test
		void shouldGetMandant() {
			service.buildSendAggregationDataResponseObserver(AggregationTestFactory.create());

			verify(properties).getMandant();
		}

		@Test
		void shouldCreateGrpcAggregationDataStream() {
			service.buildSendAggregationDataResponseObserver(AggregationTestFactory.create());

			verify(service).toGrpcAggregationDataStream(documentEntriesCaptor.capture());
			assertThat(documentEntriesCaptor.getValue()).containsExactly(AggregationTestFactory.DOCUMENT_ENTRY);
		}

		@Test
		void shouldGetBatchSize() {
			service.buildSendAggregationDataResponseObserver(AggregationTestFactory.create());

			verify(properties).getFetchingBatchSize();
		}

		@Test
		void shouldReturnResponseObserver() {
			var builtResponseObserver = service.buildSendAggregationDataResponseObserver(AggregationTestFactory.create());

			assertThat(builtResponseObserver).isNotNull();
		}

		@Nested
		class TestBuiltResponseObserver {

			private static final String MANDANT = LoremIpsum.getInstance().getWords(1);

			@BeforeEach
			void init() {
				when(properties.getMandant()).thenReturn(MANDANT);
			}

			@Test
			void shouldHaveBatchSize() {
				var builtResponseObserver = service.buildSendAggregationDataResponseObserver(AggregationTestFactory.create());

				assertThat(getBatchSize(builtResponseObserver)).isEqualTo(BATCH_SIZE);
			}

			@Test
			void shouldHaveMandant() {
				var builtResponseObserver = service.buildSendAggregationDataResponseObserver(AggregationTestFactory.create());

				assertThat(getRequestData(builtResponseObserver).mandant()).isEqualTo(MANDANT);
			}

			@Test
			void shouldHaveAggregationName() {
				var builtResponseObserver = service.buildSendAggregationDataResponseObserver(AggregationTestFactory.create());

				assertThat(getRequestData(builtResponseObserver).aggregationName()).isEqualTo(AggregationTestFactory.AGGREGATION_NAME);
			}

			@Test
			@SuppressWarnings("unchecked")
			void shouldHaveGrpcAggregationData() {
				var builtResponseObserver = service.buildSendAggregationDataResponseObserver(AggregationTestFactory.create());

				var aggregationData = IteratorUtils.toList(getRequestData(builtResponseObserver).aggregationDataIterator());
				assertThat(aggregationData).containsExactly(GrpcAggregationDataTestFactory.create());
			}

			private int getBatchSize(SendAggregationDataResponseObserver responseObserver) {
				return ReflectionTestUtils.getField(responseObserver, "batchSize", Integer.class);
			}

			private AggregationDataRemoteService.RequestData getRequestData(SendAggregationDataResponseObserver responseObserver) {
				return ReflectionTestUtils.getField(responseObserver, "requestData", AggregationDataRemoteService.RequestData.class);
			}
		}
	}

	@Nested
	class TestToGrpcAggregationDataStream {

		@BeforeEach
		void init() {
			when(grpcAggregationDataMapper.toGrpcAggregationData(AggregationTestFactory.DOCUMENT_ENTRY)).thenReturn(
					GrpcAggregationDataTestFactory.create());
		}

		@Test
		void shouldMapDocumentEntry() {
			service.toGrpcAggregationDataStream(Stream.of(AggregationTestFactory.DOCUMENT_ENTRY)).toList();

			verify(grpcAggregationDataMapper).toGrpcAggregationData(AggregationTestFactory.DOCUMENT_ENTRY);
		}

		@Test
		void shouldReturnMappedDocumentEntries() {
			var mappedDocumentEntries = service.toGrpcAggregationDataStream(Stream.of(AggregationTestFactory.DOCUMENT_ENTRY));

			assertThat(mappedDocumentEntries.toList()).containsExactly(GrpcAggregationDataTestFactory.create());
		}
	}

	@Nested
	class SendAggregationDataResponseObserverTest {

		private static final int BATCH_SIZE = 2;

		@Mock
		private AggregationDataRemoteService.RequestData requestData;
		@Spy
		private SendAggregationDataResponseObserver responseObserver = new SendAggregationDataResponseObserver(BATCH_SIZE, requestData);

		@Nested
		class TestBeforeStart {

			@Mock
			private ClientCallStreamObserver<GrpcSendAggregationDataRequest> requestObserver;
			@Mock
			private AggregationDataRemoteService.SendAggregationDataOnReadyHandler onReadyHandler;

			@BeforeEach
			void init() {
				doReturn(onReadyHandler).when(responseObserver).buildOnReadyHandler(any());
			}

			@Test
			void shouldBuildOnReadyHandler() {
				responseObserver.beforeStart(requestObserver);

				verify(responseObserver).buildOnReadyHandler(requestObserver);
			}

			@Test
			void shouldSetOnReadyHandler() {
				responseObserver.beforeStart(requestObserver);

				verify(requestObserver).setOnReadyHandler(onReadyHandler);
			}
		}

		@Nested
		class TestOnError {

			@Test
			@SneakyThrows
			void shouldCompleteExceptionally() {
				var error = new Throwable();

				responseObserver.onError(error);

				assertThatException().isThrownBy(() -> responseObserver.getResponseFuture().get()).withCause(error);
			}
		}

		@Nested
		class TestOnCompleted {

			@Test
			void shouldCompleteFuture() {
				responseObserver.onCompleted();

				assertThat(responseObserver.getResponseFuture().isDone()).isTrue();
			}
		}
	}

	@Nested
	class SendAggregationDataOnReadyHandlerTest {

		private static final int BATCH_SIZE = 2;
		private static final String MANDANT = LoremIpsum.getInstance().getWords(1);

		@Mock
		private ClientCallStreamObserver<GrpcSendAggregationDataRequest> requestObserver;

		@Nested
		class TestRun {

			@Mock
			private GrpcSendAggregationDataRequest request;

			@Test
			void shouldCheckIfStreamIsReady() {
				createOnReadyHandler(createAggregationData(1)).run();

				verify(requestObserver).isReady();
			}

			@Test
			void shouldNotCallOnNextWhenStreamIsNotReady() {
				when(requestObserver.isReady()).thenReturn(false);

				createOnReadyHandler(createAggregationData(1)).run();

				verify(requestObserver, never()).onNext(any());
			}

			@Test
			void shouldNotCallOnNextWhenHasNoMoreData() {
				when(requestObserver.isReady()).thenReturn(true);

				createOnReadyHandler(createAggregationData(0)).run();

				verify(requestObserver, never()).onNext(any());
			}

			@Test
			void shouldBuildRequest() {
				when(requestObserver.isReady()).thenReturn(true);
				var onReadyHandler = spy(createOnReadyHandler(createAggregationData(1)));

				onReadyHandler.run();

				verify(onReadyHandler).buildRequest();
			}

			@Test
			void shouldCallOnNextWithBuiltRequest() {
				when(requestObserver.isReady()).thenReturn(true).thenReturn(false);
				var onReadyHandler = spy(createOnReadyHandler(createAggregationData(1)));
				doReturn(request).when(onReadyHandler).buildRequest();

				onReadyHandler.run();

				verify(requestObserver).onNext(request);
			}

			@Test
			void shouldCallOnNextUntilAllDataWasSent() {
				when(requestObserver.isReady()).thenReturn(true);

				createOnReadyHandler(createAggregationData(BATCH_SIZE + 1)).run();

				verify(requestObserver, times(2)).onNext(any());
			}

			@Test
			void shouldCompleteRequest() {
				when(requestObserver.isReady()).thenReturn(true);

				createOnReadyHandler(createAggregationData(BATCH_SIZE + 1)).run();

				verify(requestObserver).onCompleted();
			}

			@Test
			void shouldNotCompleteRequestIfNotAllDataWasSent() {
				when(requestObserver.isReady()).thenReturn(true).thenReturn(false);

				createOnReadyHandler(createAggregationData(BATCH_SIZE + 1)).run();

				verify(requestObserver, never()).onCompleted();
			}
		}

		@Nested
		class TestBuildRequest {

			@Test
			void shouldSetName() {
				var request = createOnReadyHandler().buildRequest();

				assertThat(request.getName()).isEqualTo(AggregationTestFactory.AGGREGATION_NAME);
			}

			@Test
			void shouldSetMandant() {
				var request = createOnReadyHandler().buildRequest();

				assertThat(request.getMandant()).isEqualTo(MANDANT);
			}

			@Nested
			class OnHasNoAggregationData {

				@Test
				void shouldAggregationDataBeEmpty() {
					var request = createOnReadyHandler(List.of()).buildRequest();

					assertThat(request.getAggregationDataCount()).isEqualTo(0);
				}
			}

			@Nested
			class OnHasLessAggregationDataThenBatchSize {

				@Test
				void shouldAddAllAggregationData() {
					var aggregationData = createAggregationData(BATCH_SIZE - 1);

					var request = createOnReadyHandler(aggregationData).buildRequest();

					assertThat(request.getAggregationDataList()).hasSameElementsAs(aggregationData);
				}
			}

			@Nested
			class OnHasBatchSizeOfAggregationData {

				@Test
				void shouldAddAllAggregationData() {
					var aggregationData = createAggregationData(BATCH_SIZE);

					var request = createOnReadyHandler(aggregationData).buildRequest();

					assertThat(request.getAggregationDataList()).hasSameElementsAs(aggregationData);
				}
			}

			@Nested
			class OnHasMoreAggregationDataThenBatchSize {

				@Test
				void shouldHaveExactlyBatchSizeOfAggregationData() {
					var aggregationData = createAggregationData(BATCH_SIZE + 1);

					var request = createOnReadyHandler(aggregationData).buildRequest();

					assertThat(request.getAggregationDataCount()).isEqualTo(BATCH_SIZE);
				}

				@Test
				void shouldContainOnlyFirstBatchSizeOfElements() {
					var aggregationData = createAggregationData(BATCH_SIZE + 1);

					var request = createOnReadyHandler(aggregationData).buildRequest();

					assertThat(request.getAggregationDataList()).containsExactlyElementsOf(aggregationData.subList(0, BATCH_SIZE));
				}
			}
		}

		private List<GrpcAggregationData> createAggregationData(int count) {
			return IntStream.range(1, count + 1)
					.mapToObj(idx -> GrpcAggregationDataTestFactory.createBuilder().setVorgangName("vorgang " + idx).build())
					.toList();
		}

		private AggregationDataRemoteService.SendAggregationDataOnReadyHandler createOnReadyHandler() {
			return createOnReadyHandler(List.of());
		}

		private AggregationDataRemoteService.SendAggregationDataOnReadyHandler createOnReadyHandler(List<GrpcAggregationData> aggregationData) {
			return new AggregationDataRemoteService.SendAggregationDataOnReadyHandler(requestObserver, BATCH_SIZE,
					AggregationDataRemoteService.RequestData.builder()
							.mandant(MANDANT)
							.aggregationName(AggregationTestFactory.AGGREGATION_NAME)
							.aggregationDataIterator(aggregationData.iterator())
							.build());
		}
	}
}
