/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.aggregation.warehouse;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.aggregation.Aggregation;
import de.ozgcloud.aggregation.AggregationTestFactory;
import de.ozgcloud.aggregation.transformation.AggregationMapping;

class AggregationDataWarehouseLoaderTest {

	@Mock
	private WarehouseRepository repository;
	@Spy
	@InjectMocks
	private AggregationDataWarehouseLoader loader;

	@Nested
	class TestLoadIntoTarget {

		private static final String COLLECTION_NAME = LoremIpsum.getInstance().getWords(1);
		private final Aggregation aggregation = AggregationTestFactory.create();

		@Captor
		private ArgumentCaptor<Stream<DocumentEntry>> documentEntriesCaptor;

		@BeforeEach
		void init() {
			doReturn(COLLECTION_NAME).when(loader).getCollectionName(aggregation);
		}

		@Test
		void shouldGetCollectionName() {
			loader.loadIntoTarget(aggregation);

			verify(loader).getCollectionName(aggregation);
		}

		@Test
		void shouldClearCollection() {
			loader.loadIntoTarget(aggregation);

			verify(repository).clearCollection(COLLECTION_NAME);
		}

		@Test
		void shouldSaveDocumentEntriesInCollection() {
			loader.loadIntoTarget(aggregation);

			verify(repository).saveAllInCollection(documentEntriesCaptor.capture(), eq(COLLECTION_NAME));
			assertThat(documentEntriesCaptor.getValue()).containsExactly(AggregationTestFactory.DOCUMENT_ENTRY);
		}
	}

	@Nested
	class TestGetCollectionName {

		@Test
		void shouldReturnAggregationName() {
			var collectionName = loader.getCollectionName(AggregationTestFactory.create());

			assertThat(collectionName).isEqualTo(AggregationTestFactory.AGGREGATION_NAME);
		}

		@ParameterizedTest
		@NullAndEmptySource
		void shouldReturnCollectionNameFromDocumentEntry(String aggregationName) {
			var collectionName = loader.getCollectionName(AggregationTestFactory.createBuilder().aggregationName(aggregationName).build());

			assertThat(collectionName).isEqualTo(DocumentEntry.COLLECTION);
		}
	}

	@Nested
	class TestGetScope {

		@Test
		void shouldReturnInternScope() {
			var scope = loader.getScope();

			assertThat(scope).isEqualTo(AggregationMapping.Scope.INTERN);
		}
	}
}
